@@ -1,9 +1,9 @@
 #pragma once
 
 #include <stdbool.h>
-#include <sys/types.h>
 
 #include <osmocom/core/logging.h>
+#include <osmocom/core/thread.h>
 
 extern const struct log_info log_info;
 
@@ -16,14 +16,13 @@ enum {
 	DTRXDUL,
 	DDEV,
 	DDEVDRV,
+	DCTR,
 };
 
-pid_t my_gettid(void);
-
 #define CLOGC(category, level, fmt, args...) do { \
-	LOGP(category, level, "[tid=%ld] " fmt, (long int) my_gettid(), ##args);  \
+	LOGP(category, level, "[tid=%ld] " fmt, (long int) osmo_gettid(), ##args);  \
 } while(0)
 
 #define CLOGCHAN(chan, category, level, fmt, args...) do { \
-	LOGP(category, level, "[tid=%ld][chan=%zu] " fmt, (long int) my_gettid(), chan, ##args);  \
+	LOGP(category, level, "[tid=%ld][chan=%zu] " fmt, (long int) osmo_gettid(), chan, ##args);  \
 } while(0)
